package org.supermind.crawl;

import java.net.URL;

/**
 * A data structure for managing the list of URLs to fetch. Roughly equivalent
 * to Mercator's Frontier. This class is NOT thread-safe.
 * Each FetcherThread is supposed to have its own fetchlist.
 */
public interface FetchList {

  /**
   * Initialize resources.
   */
  void init();

  /**
   * Get next {@link HostQueue}.
   * @return next HostQueue, or null if none of the HostQueues have any URLs
   */
  HostQueue next();

  /**
   * Release HostQueue from use. This method must be called when
   * {@link HostQueue#pop()} completes.
   *
   * @param hostQueue
   * @param popped number of URLs popped from the queue
   * @param timeTaken total time taken to download the popped urls
   */
  void release(HostQueue hostQueue, int popped, long timeTaken);

  /**
   * Add a {@link ScheduledURL} to the fetchlist. Multiple threads
   * can be calling this method, and implementing classes must
   * synchronize access accordingly.
   * @param parent originating url
   * @param url url to queue
   */
  void queue(ScheduledURL parent, URL url);

  /**
   * Does the fetchlist contain this url?
   * @param url
   * @return
   */
  boolean contains(URL url);

  /**
   * Total number of URLs this fetchlist currently contains. This variable is
   * only updated when HostQueues have been {@link #release}d.
   * @return number of URLs
   */
  int getCurrentSize();

  /**
   * Release resources.
   */
  void close();
}
