package org.supermind.crawl;

import org.apache.nutch.util.LogFormatter;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.BeanFactoryAware;

import java.io.IOException;
import java.net.URL;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * Fetcher.
 */
public class Fetcher implements BeanFactoryAware {
  /**
   * Logger.
   */
  protected static Logger LOG =
      LogFormatter.getLogger(Fetcher.class.getName());

  // All threads (FetcherThread or thread started by it) belong to
  // group "fetcher". Each FetcherThread is named as "fetcherXX",
  // where XX is the order it's started.
  protected static final String THREAD_GROUP_NAME = "fetcher";

  protected final ThreadGroup group = new ThreadGroup(THREAD_GROUP_NAME); // our group

  protected boolean parsing;
  protected int threadCount;
  protected BeanFactory beanFactory;
  protected CrawlSeedSource crawlSeedSource;
  protected PostFetchProcessor postFetchProcessor = PostFetchProcessor.NO_OP;

  protected long start;                             // start time of fetcher run
  protected long bytes;                             // total bytes fetched
  protected int pages;                              // total pages fetched
  protected int errors;                             // total pages errored

  /**
   * The maximum number of pipedlined HTTP GETs to perform per HTTP connection.
   */
  protected int maxPagesPerConnection = 5;


  public Fetcher() throws Exception {
  }

  public void run() throws IOException, InterruptedException {
    init();
    // Quit monitoring if all FetcherThreads are gone.
    // There could still be other threads, which may well be runaway threads
    // started by external libs via FetcherThreads and it is generally safe
    // to ignore them because our main FetcherThreads have finished their jobs.
    // In fact we are a little more cautious here by making sure
    // there is no more outstanding page fetches via monitoring
    // changes of pages, errors and bytes.
    int pages0 = pages;
    int errors0 = errors;
    long bytes0 = bytes;

    while (true) {
      Thread.sleep(1000);

      if (LogFormatter.hasLoggedSevere())
        throw new RuntimeException("SEVERE error logged.  Exiting fetcher.");

      int n = group.activeCount();
      Thread[] list = new Thread[n];
      group.enumerate(list);
      boolean noMoreFetcherThread = true; // assumption
      boolean allWaiting = true;
      for (int i = 0; i < n; i++) {
        // this thread may have gone away in the meantime
        if (list[i] == null) continue;
        String tname = list[i].getName();
        if (tname.startsWith(THREAD_GROUP_NAME)) // prove it
          noMoreFetcherThread = false;
        if (!((FetcherThread) list[i]).waiting)
          allWaiting = false;
        if (LOG.isLoggable(Level.FINEST))
          LOG.finest(list[i].toString() + " " + tname);
      }

      if (allWaiting) {
        for (int i = 0; i < n; i++) {
          if (list[i] == null) continue;
          ((FetcherThread) list[i]).continueWaiting = false;
        }
      }

      if (noMoreFetcherThread) {
        if (LOG.isLoggable(Level.FINE))
          LOG.fine("number of active threads: " + n);
        if (pages == pages0 && errors == errors0 && bytes == bytes0)
          break;
        status();
        pages0 = pages;
        errors0 = errors;
        bytes0 = bytes;
      }
    }
    close();
  }

  protected void close() throws IOException {
    postFetchProcessor.close();
  }

  /**
   * Display the status of the fetcher run.
   */
  public synchronized void status() {
    org.apache.nutch.fetcher.Fetcher.FetcherStatus status = getStatus();
    LOG.info(status.toString());
    LOG.info("status: "
        + (((float) status.getPageCount()) / (status.getElapsedTime() / 1000.0f)) + " pages/s, "
        + (((float) status.getByteCount() * 8 / 1024) / (status.getElapsedTime() / 1000.0f)) + " kb/s, "
        + (((float) status.getByteCount()) / status.getPageCount()) + " bytes/page");
  }

  public synchronized org.apache.nutch.fetcher.Fetcher.FetcherStatus getStatus() {
    return new org.apache.nutch.fetcher.Fetcher.FetcherStatus("", start, pages, errors, bytes);
  }

  private void init() throws IOException {
    this.start = System.currentTimeMillis();
    if (LOG.isLoggable(Level.FINER))
      LOG.finer("Initializing " + threadCount + " threads");
    FetcherThread[] threads = new FetcherThread[threadCount];
    for (int i = 0; i < threadCount; i++) {       // spawn threads
      FetcherThread thread = (FetcherThread) beanFactory.getBean("fetcherThread");
      thread.setDaemon(true);
      thread.start();
      threads[i] = thread;
    }

    for (Iterator<SeedURL> it = crawlSeedSource.getSeedURLs(); it.hasNext();) {
      SeedURL seed = it.next();
      if (LOG.isLoggable(Level.FINER)) LOG.finer("Added seed URL:" + seed.url);
      FetcherThread thread = assignThread(seed.url);
      thread.fetchList.queue(new ScheduledURL(seed.idx, -1, null), seed.url);
    }
    crawlSeedSource.close();
  }

  /**
   * Assign a fetcherthread to handle this URL.
   * TODO: implement a more sophisticated assignment algorithm (which load-balances for e.g.).
   *
   * @param url
   * @return
   */
  protected FetcherThread assignThread(final URL url) {
    Thread[] threads = new Thread[group.activeCount()];
    group.enumerate(threads);
    int threadNumber = Math.abs(url.getHost().hashCode() % threads.length);
    if (LOG.isLoggable(Level.FINEST))
      LOG.finest("Assigning " + url + " to thread: "
          + threads[threadNumber].getName() + "/" + threads.length);
    FetcherThread fetcherThread = (FetcherThread) threads[threadNumber];
    return fetcherThread;
  }

  public void setSeedSource(CrawlSeedSource crawlSource) throws IOException {
    this.crawlSeedSource = crawlSource;
  }

  public void setPostFetchProcessor(PostFetchProcessor postFetchProcessor) {
    this.postFetchProcessor = postFetchProcessor;
  }

  public void setBeanFactory(BeanFactory beanFactory) {
    this.beanFactory = beanFactory;
  }

  public void setThreadCount(int threadCount) {
    this.threadCount = threadCount;
  }

  public void setParsing(boolean parsing) {
    this.parsing = parsing;
  }

  public void setMaxPagesPerConnection(int maxPagesPerConnection) {
    this.maxPagesPerConnection = maxPagesPerConnection;
  }
}
