package org.supermind.crawl;

import org.apache.nutch.io.MD5Hash;
import org.apache.nutch.parse.*;
import org.apache.nutch.protocol.Content;
import org.apache.nutch.protocol.Protocol;
import org.apache.nutch.protocol.ProtocolOutput;
import org.apache.nutch.protocol.ProtocolStatus;
import org.supermind.crawl.http.Http;
import org.supermind.crawl.scope.FetchListScope;
import org.supermind.crawl.scope.ParseScope;
import org.supermind.crawl.scope.PostFetchScope;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Thread that performs the actual fetching.
 * <p/>
 * Each FetcherThread has its own {@link FetchList} and {@link FetchedURLs}.
 * URLs are assigned to FetcherThreads by host, and the same URL is
 * guaranteed to be always assigned to the same host.
 * </p>
 * <p/>
 * URLs queued for fetching are not added directly to the respective fetchlists,
 * rather to a holding area, which is checked at regular intervals. This allows
 * fetchlists and fetchedurls to operate in a single-threaded model.
 * </p>
 */
public class FetcherThread extends Thread {
  protected final Logger LOG = Fetcher.LOG;
  protected FetchList fetchList;
  protected PostFetchScope postFetchScope;
  protected FetchListScope fetchListScope;
  protected ParseScope parseScope;
  protected FetchedURLs fetchedUrls;
  protected Fetcher fetcher;

  protected final FetchListScope.Input flScopeIn = new FetchListScope.Input();
  protected final PostFetchScope.Input postFetchInput = new PostFetchScope.Input();

  protected boolean continueWaiting = true;
  protected boolean waiting;

  protected LinkedHashMap<URL, ScheduledURL> linkQueue = new LinkedHashMap();

  protected int linkQueueBatchSize = 10;

  public FetcherThread(Fetcher fetcher) {
    super(fetcher.group, Fetcher.THREAD_GROUP_NAME + String.valueOf(fetcher.group.activeCount()));
    this.fetcher = fetcher;
  }

  public void run() {
    try {
      this.fetchList.init();
      this.fetchedUrls.init();
    } catch (IOException e) {
      throw new RuntimeException(e);
    }

    HostQueue next = null;
    while (true) {
      waiting = (next = fetchList.next()) == null;

      // pop batch URLs from linkQueue to add to fetchList
      if (linkQueue.size() > linkQueueBatchSize || waiting) {
        synchronized (linkQueue) {
          int i = 0;
          for (Iterator<URL> it = linkQueue.keySet().iterator();
               it.hasNext() && i < linkQueueBatchSize; ++i) {
            URL url = it.next();
            ScheduledURL parent = linkQueue.get(url);
            it.remove();

            if (fetchedUrls.contains(url)
                || fetchList.contains(url))
              continue;
            flScopeIn.fill(parent, url);
            if (fetchListScope.isInScope(flScopeIn)) {
              fetchList.queue(parent, url);
            }
            flScopeIn.clear();
          }
        }
      }

      if (waiting) {
        if (continueWaiting) {
          try {
            Thread.sleep(500);
          } catch (InterruptedException e) {
          }
          continue;
        } else {
          break;
        }
      }
      if (LOG.isLoggable(Level.FINER))
        LOG.finer("Popped hostqueue:" + next);
      long start = System.currentTimeMillis();
      ScheduledURL[] pages = next.pop();
      Protocol protocol = null;
      try {
        if (next.getProtocol().equals("http")) {
          protocol = new Http();
          ProtocolOutput[] outputs = ((Http) protocol).getProtocolOutput(pages);
          // log the fetched urls first so no duplicate links added
          for (int i = 0; i < outputs.length; i++) {
//            LOG.finer("Inserting fetched URL:" + pages[i].url);
            fetchedUrls.insert(pages[i], outputs[i]);
          }

          for (int i = 0; i < outputs.length; i++) {
            handleFetch(pages[i], outputs[i]);
          }
          // FIXME: non-HTTP protocols aren't handled for now
//        } else {
//          protocol = ProtocolFactory.getProtocolByName(next.getProtocol());
//          LOG.fine("Fetching non-http hostqueue:" + next);
//          for (ScheduledURL scheduledURL : pages) {
//            ProtocolOutput output = protocol.getProtocolOutput(scheduledURL.url.toString());
//            // log the fetched urls first so no duplicate links added
//            fetchedUrls.insert(scheduledURL, output);
//            handleFetch(scheduledURL, output);
//          }
        }
//      } catch (ProtocolNotFound protocolNotFound) {
      } catch (IOException e) {
        LOG.log(Level.SEVERE, e.getMessage(), e);
      } finally {
        long stop = System.currentTimeMillis();
        fetchList.release(next, pages.length, (stop - start));
      }
    }
    try {
      close();
    } catch (IOException e) {
      LOG.log(Level.SEVERE, e.getMessage(), e);
    }
  }

  private void close() throws IOException {
    fetchedUrls.close();
    fetchList.close();
  }

  protected ParseStatus handleFetch(ScheduledURL scheduledURL, ProtocolOutput output) throws IOException {
    Content content = output.getContent();
    MD5Hash hash = null;
    String urlstring = scheduledURL.url.toString();

    if (content == null) {
      content = new Content(urlstring, urlstring, new byte[0], "", new Properties());
      hash = MD5Hash.digest(urlstring);
    } else {
      hash = MD5Hash.digest(content.getContent());
    }
    ProtocolStatus protocolStatus = output.getStatus();

    switch (protocolStatus.getCode()) {
      case ProtocolStatus.SUCCESS:
        if (content != null) {
          // more accurate reflection of actual bytes downloaded for gzipped content
          String headerLength = content.getMetadata().getProperty("Content-Length");
          long length = headerLength != null ? Long.parseLong(headerLength)
              : content.getContent().length;

          synchronized (fetcher) {           // update status
            ++fetcher.pages;
            fetcher.bytes += length;
            if ((fetcher.pages % 100) == 0) {             // show status every 100pp
              fetcher.status();
            }
          }
        }
        break;
      case ProtocolStatus.EXCEPTION:
        synchronized (fetcher) {
          LOG.warning("ERR:" + urlstring + " " + output.getStatus());
          ++fetcher.errors;
        }
    }

    final FetcherOutput fo = new FetcherOutput(scheduledURL, hash, protocolStatus);

    // parse
    ParseStatus parseStatus = null;
    Parse parse = null;
    if (parseScope.isInScope(fo)) {
      String contentType = content.getContentType();
      Parser parser = null;
      try {
        parser = ParserFactory.getParser(contentType, urlstring);
        parse = parser.getParse(content);
        parseStatus = parse.getData().getStatus();
      } catch (Exception e) {
        LOG.log(Level.WARNING, e.getMessage(), e);
        parseStatus = new ParseStatus(e);
      }
    }

    // output
    if (parseStatus.isSuccess()) {
      /**
       * Handle content-level META redirect. TODO: count number of redirects
       */
      if (parseStatus.getMinorCode() == ParseStatus.SUCCESS_REDIRECT) {
        final String toUrl = parseStatus.getMessage();
        queueOutlink(scheduledURL, toUrl);
      } else {
        postFetchInput.fill(fo, parse);
        if (postFetchScope.isInScope(postFetchInput)) {
          fetcher.postFetchProcessor.process(fo, content, parse);
        }
        postFetchInput.clear();
        addOutlinksToFetchlist(scheduledURL, parse);
      }
    } else {
      LOG.info("fetch okay, but can't parse " + urlstring + ", reason: "
          + parseStatus.toString());
      postFetchInput.fill(fo, parse);
      if (postFetchScope.isInScope(postFetchInput)) {
        fetcher.postFetchProcessor.process(fo, content, parse);
      }
      postFetchInput.clear();
    }

    return parseStatus;
  }

  /**
   * Add outlinks to fetchlist.
   *
   * @param parent
   * @param parse
   */
  protected void addOutlinksToFetchlist(ScheduledURL parent, Parse parse) {
    if (parse == null) return;

    ParseData data = parse.getData();
    Outlink[] outlinks = data.getOutlinks();
    for (int i = 0; i < outlinks.length; ++i) {
      final String toUrl = outlinks[i].getToUrl();
      queueOutlink(parent, toUrl);
      continue;
    }

  }

  private void queueOutlink(ScheduledURL parent, final String toUrl) {
    final URL url;
    FetcherThread fetcherThread;
    try {
      url = new URL(toUrl);
      fetcherThread = fetcher.assignThread(url);
      synchronized (fetcherThread.linkQueue) {
        fetcherThread.linkQueue.put(url, parent);
      }
    } catch (MalformedURLException e) {
      return;
    }
  }

  public void setFetcher(Fetcher fetcher) {
    this.fetcher = fetcher;
  }

  public void setFetchList(FetchList fetchList) {
    this.fetchList = fetchList;
  }

  public void setPostFetchScope(PostFetchScope postFetchScope) {
    this.postFetchScope = postFetchScope;
  }

  public void setFetchListScope(FetchListScope fetchListScope) {
    this.fetchListScope = fetchListScope;
  }

  public void setParseScope(ParseScope parseScope) {
    this.parseScope = parseScope;
  }

  public void setFetchedURLs(FetchedURLs fetchedUrls) {
    this.fetchedUrls = fetchedUrls;
  }



//    protected ParseStatus handleFetch(FetchListEntry fle, ProtocolOutput output) {
//      Content content = output.getContent();
//      MD5Hash hash = null;
//      String url = fle.getPage().getURL().toString();
//
//      fetchedUrls.insert(url);
//
//      if (content == null) {
//        content = new Content(url, url, new byte[0], "", new Properties());
//        hash = MD5Hash.digest(url);
//      } else {
//        hash = MD5Hash.digest(content.getContent());
//      }
//      ProtocolStatus protocolStatus = output.getStatus();
////      if (!this.parsing) {
////        pageOutputter.outputPage(new FetcherOutput(fle, hash, protocolStatus),
////                content, null, null);
////        return null;
////      }
//
//      final FetcherOutput fo = new FetcherOutput(fle, hash, protocolStatus);
//      ParseStatus status = null;
//      if (parseScope.filter(fo)) {
//        String contentType = content.getContentType();
//        Parser parser = null;
//        Parse parse = null;
//        try {
//          parser = ParserFactory.getParser(contentType, url);
//          parse = parser.getParse(content);
//          status = parse.getData().getStatus();
//        } catch (Exception e) {
//          e.printStackTrace();
//          status = new ParseStatus(e);
//        }
//
//        outputScopeIn.fill(fo, parse);
//        if (status.isSuccess() && postFetchScope.filter(outputScopeIn)) {
//          pageOutputter.outputPage(fo,
//              content, new ParseText(parse.getText()), parse.getData());
//          addOutlinksToFetchlist(fle, parse.getData());
//        }
//        outputScopeIn.clear();
//      }
////    } else {
////      LOG.info("fetch okay, but can't parse " + url + ", reason: "
////          + status.toString());
////      pageOutputter.outputPage(fo,
////          content, new ParseText(""),
////          new ParseData(status, "", new Outlink[0], new Properties()));
////    }
//
//      return status;
//    }


}   // end FetcherThread
