package org.supermind.crawl.http;

import org.apache.nutch.db.Page;
import org.apache.nutch.pagedb.FetchListEntry;
import org.apache.nutch.protocol.*;
import org.apache.nutch.util.LogFormatter;
import org.apache.nutch.util.NutchConf;
import org.supermind.crawl.ScheduledURL;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.LinkedList;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Takes advantage of HTTP 1.1 features such as connection persistence and
 * request pipelining to improve performance for situations where multiple URLs
 * must be fetched from the same host.
 */
public class Http implements Protocol {
  public static final Logger LOG =
      LogFormatter.getLogger(Http.class.getName());

  static {
    if (NutchConf.get().getBoolean("http.verbose", false))
      LOG.setLevel(Level.FINE);
  }

  static final int BUFFER_SIZE = 8 * 1024;

  static final int MAX_REDIRECTS =
      NutchConf.get().getInt("http.redirect.max", 3);

  static String PROXY_HOST = NutchConf.get().get("http.proxy.host");
  static int PROXY_PORT = NutchConf.get().getInt("http.proxy.port", 8080);
  static boolean PROXY = (PROXY_HOST != null && PROXY_HOST.length() > 0);

  static int TIMEOUT = NutchConf.get().getInt("http.timeout", 10000);
  static int MAX_CONTENT = NutchConf.get().getInt("http.content.limit", 64 * 1024);

  static int MAX_DELAYS = NutchConf.get().getInt("http.max.delays", 3);

  static String AGENT_STRING = getAgentString();

  static long SERVER_DELAY =
      (long) (NutchConf.get().getFloat("fetcher.server.delay", 1.0f) * 1000);

  static {
    LOG.info("http.proxy.host = " + PROXY_HOST);
    LOG.info("http.proxy.port = " + PROXY_PORT);

    LOG.info("http.timeout = " + TIMEOUT);
    LOG.info("http.content.limit = " + MAX_CONTENT);
    LOG.info("http.agent = " + AGENT_STRING);

    LOG.info("fetcher.server.delay = " + SERVER_DELAY);
    LOG.info("http.max.delays = " + MAX_DELAYS);
  }

  private boolean bypassRobots = false;

  public ProtocolOutput getProtocolOutput(String urlString) {
    try {
      return getProtocolOutput(new FetchListEntry(true,
          new Page(urlString, 1.0f), new String[0]));
    } catch (MalformedURLException mue) {
      return new ProtocolOutput(null, new ProtocolStatus(mue));
    }
  }

  public ProtocolOutput[] getProtocolOutput(String[] urlstrings) {
    ProtocolOutput[] result = new ProtocolOutput[urlstrings.length];
    LinkedList<URL> urlList = new LinkedList();
    for (int i = 0; i < urlstrings.length; ++i) {
      try {
        URL url = new URL(urlstrings[i]);
        if (!bypassRobots && !RobotRulesParser.isAllowed(url)) {
          result[i] = new ProtocolOutput(null, new ProtocolStatus(ProtocolStatus.ROBOTS_DENIED, url));
          continue;
        }
        urlList.add(url);
      } catch (Throwable e) {
        // XXX Maybe bogus: assume this is allowed.
        LOG.fine("Exception checking robot rules for " + urlstrings[i] + ": " + e);
      }
    }
    URL[] urls = urlList.toArray(new URL[0]);
    return getProtocolOutput(urls, result);
  }

  public ProtocolOutput[] getProtocolOutput(ScheduledURL[] pages) {
    ProtocolOutput[] result = new ProtocolOutput[pages.length];
    LinkedList<URL> urlList = new LinkedList();
    for (int i = 0; i < pages.length; ++i) {
      URL url = pages[i].url;
      try {
        if (!bypassRobots && !RobotRulesParser.isAllowed(url)) {
          result[i] = new ProtocolOutput(null, new ProtocolStatus(ProtocolStatus.ROBOTS_DENIED, url));
          continue;
        }
      } catch (Throwable e) {
        // XXX Maybe bogus: assume this is allowed.
        LOG.fine("Exception checking robot rules for " + url + ": " + e);
      }
      urlList.add(url);
    }
    URL[] urls = urlList.toArray(new URL[0]);
    return getProtocolOutput(urls, result);
  }

  /**
   * @param urls
   * @param result
   * @return
   */
  private ProtocolOutput[] getProtocolOutput(URL[] urls, ProtocolOutput[] result) {
    HttpResponse response = new HttpResponse(); // make a request
    Content[] contents = null;
//    try {
      contents = response.get(null, urls);
//    } catch (Throwable t) {
//      LOG.log(Level.WARNING, "Error in getting response", t);
//      for (int i = 0; i < result.length; ++i) {
//        result[i] = new ProtocolOutput(null, new ProtocolStatus(t));
//      }
//      return result;
//    }

    for (int i = 0, j = 0; i < contents.length; ++i, ++j) {
      try {
        Content c = contents[i];
        if(c == null) {
          result[j] = new ProtocolOutput(null, ProtocolStatus.STATUS_FAILED);
          continue;
        }
        URL url = new URL(c.getUrl());

        int code = response.getCode(url.toString());

        while(result[j] != null) ++j; // some elements of result[] may be filled because of robot rules

        if (code == 200) { // got a good response
          result[j] = new ProtocolOutput(c);

        } else if (code == 410) { // page is gone
          result[j] = new ProtocolOutput(c, new ProtocolStatus(ProtocolStatus.GONE, "Http: " + code + " url=" + url));
        } else if (code == ProtocolStatus.REDIR_EXCEEDED) {
          /**
           * Hack, making use of the fact that all HTTP status codes > 100, so
           * as long as ProtocolStatus' constants are < 100, its ok. There has to be
           * some way for HttpResponse to communicate that redir_exceeded, without
           * throwing exception.
           */
          result[j] = new ProtocolOutput(c, new ProtocolStatus(ProtocolStatus.REDIR_EXCEEDED,
              "Too many redirects: " + url.toString()));
        } else if (code >= 300 && code < 400) { // handle redirect
          /**
           * Actual redirection (i.e. 302) is handled in HttpResponse
           * If we get here, it means codes like 300 were returned, to
           * which I don't really know what to do with it, so just cascade
           * this to the higher layer.
           */
          int protocolStatusCode;
          switch (code) {
            case 300:   // multiple choices, preferred value in Location
              protocolStatusCode = ProtocolStatus.MOVED;
              break;
            case 301:   // moved permanently
            case 305:   // use proxy (Location is URL of proxy)
              protocolStatusCode = ProtocolStatus.MOVED;
              break;
            case 302:   // found (temporarily moved)
            case 303:   // see other (redirect after POST)
            case 307:   // temporary redirect
              protocolStatusCode = ProtocolStatus.TEMP_MOVED;
              break;
            case 304:   // not modified
              protocolStatusCode = ProtocolStatus.NOTMODIFIED;
              break;
            default:
              protocolStatusCode = ProtocolStatus.MOVED;
          }
          // handle this in the higher layer.
          result[j] = new ProtocolOutput(c, new ProtocolStatus(protocolStatusCode, url));
        } else if (code == 400) { // bad request, mark as GONE
          LOG.fine("400 Bad request: " + url);
          result[j] = new ProtocolOutput(c, new ProtocolStatus(ProtocolStatus.GONE, url));
        } else if (code == 401) { // requires authorization
          LOG.fine("401 Authentication Required");
//          Properties p = c.getMetadata();
//            if (p instanceof MultiProperties) {
//              auth = HttpAuthenticationFactory.findAuthentication((MultiProperties) p);
//            } else {
          result[j] = new ProtocolOutput(c, new ProtocolStatus(ProtocolStatus.ACCESS_DENIED, "Authorization required: "
              + url.toString()));
//            }
        } else if (code == 404) {
          result[j] = new ProtocolOutput(c, new ProtocolStatus(ProtocolStatus.NOTFOUND, url));
        } else if (code == 410) { // permanently GONE
          result[j] = new ProtocolOutput(c, new ProtocolStatus(ProtocolStatus.GONE, url));
        } else {
          result[j] = new ProtocolOutput(c, new ProtocolStatus(ProtocolStatus.EXCEPTION, "Http code=" + code + ", url="
              + url));
        }
      } catch (Throwable e) {
        LOG.log(Level.WARNING, "Exception creating protocoloutput", e);
        result[j] = new ProtocolOutput(null, new ProtocolStatus(e));
      }
    }

    return result;
  }

  public ProtocolOutput getProtocolOutput(FetchListEntry fle) {
    String urlString = fle.getUrl().toString();
    try {
      URL url = new URL(urlString);

      int redirects = 0;
      while (true) {

        if (!RobotRulesParser.isAllowed(url))
          throw new ResourceGone(url, "Blocked by robots.txt");

        HttpResponse response;
        response = new HttpResponse(); // make a request
        Content[] contents = response.get(null, new URL[]{url});
        int code = response.getCode(url.toString());

        if (code == 200) {                        // got a good response
          return new ProtocolOutput(contents[0]);            // return it

        } else if (code == 410) {                 // page is gone
          throw new ResourceGone(url, "Http: " + code);

        } else if (code >= 300 && code < 400) {   // handle redirect
          if (redirects == MAX_REDIRECTS)
            throw new HttpException("Too many redirects: " + urlString);
          url = new URL(url, contents[0].getMetadata().getProperty("Location"));
          redirects++;
          LOG.fine("redirect to " + url);

        } else {                                  // convert to exception
          throw new HttpError(code);
        }
      }
    } catch (Exception e) {
      return new ProtocolOutput(null, new ProtocolStatus(e));
    }
  }

  private static String getAgentString() {
    String agentName = NutchConf.get().get("http.agent.name");
    String agentVersion = NutchConf.get().get("http.agent.version");
    String agentDesc = NutchConf.get().get("http.agent.description");
    String agentURL = NutchConf.get().get("http.agent.url");
    String agentEmail = NutchConf.get().get("http.agent.email");

    if ((agentName == null) || (agentName.trim().length() == 0))
      LOG.severe("No User-Agent string set (http.agent.name)!");

    StringBuffer buf = new StringBuffer();

    buf.append(agentName);
    if (agentVersion != null) {
      buf.append("/");
      buf.append(agentVersion);
    }
    if (((agentDesc != null) && (agentDesc.length() != 0))
        || ((agentEmail != null) && (agentEmail.length() != 0))
        || ((agentURL != null) && (agentURL.length() != 0))) {
      buf.append(" (");

      if ((agentDesc != null) && (agentDesc.length() != 0)) {
        buf.append(agentDesc);
        if ((agentURL != null) || (agentEmail != null))
          buf.append("; ");
      }

      if ((agentURL != null) && (agentURL.length() != 0)) {
        buf.append(agentURL);
        if (agentEmail != null)
          buf.append("; ");
      }

      if ((agentEmail != null) && (agentEmail.length() != 0))
        buf.append(agentEmail);

      buf.append(")");
    }
    return buf.toString();
  }

  /**
   * For debugging.
   */
  public static void main(String[] args) throws Exception {
    boolean verbose = true;
    String url = null;
    String usage = "Usage: Http [-verbose] [-timeout N] url";

    if (args.length == 0) {
      System.err.println(usage);
      System.exit(-1);
    }

    for (int i = 0; i < args.length; i++) {       // parse command line
      if (args[i].equals("-timeout")) {           // found -timeout option
        TIMEOUT = Integer.parseInt(args[++i]) * 1000;
      } else if (args[i].equals("-verbose")) {    // found -verbose option
        verbose = true;
      } else if (i != args.length - 1) {
        System.err.println(usage);
        System.exit(-1);
      } else                                      // root is required parameter
        url = args[i];
    }

    Http http = new Http();

    if (verbose) {
      LOG.setLevel(Level.FINER);
    }

    Content content = http.getProtocolOutput(url).getContent();

    System.out.println("Content Type: " + content.getContentType());
    System.out.println("Content Length: " + content.get("Content-Length"));
    System.out.println("Content:");
    String text = new String(content.getContent());
    System.out.println(text);

  }

}
