package org.supermind.crawl.util;

import org.apache.nutch.util.LogFormatter;

import java.io.IOException;
import java.util.logging.Logger;

/**
 * A fixed-size chained scatter table that implements the clock/second-chance
 * algorithm for purging entries.
 * <p/>
 * Implemented for storing longs. The 16 most significant bits are used as the
 * hashkey, whilst the 48 least significant bits are used as the value.
 */
public class ChainedScatterTable {
  /**
   * Logger.
   */
  protected static Logger LOG =
      LogFormatter.getLogger(ChainedScatterTable.class.getName());

  protected Entry[] array = new Entry[1 << 16]; // 65536
  protected int count;
  protected int clockIndex = 0;
  protected ScatterPurger purger;

  public ChainedScatterTable(ScatterPurger purger) {
    for (int i = 0; i < array.length; ++i) {
      array[i] = new Entry();
    }
    this.purger = purger;
  }

  public void clear() {
    for (int i = 0; i < array.length; ++i) {
      array[i].purge();
    }
    clockIndex = 0;
  }

  public int insert(long value) {
    /**
     * 4 possible scenarions
     * 1. Not full, no collision
     * 2. Not full, collision
     * 3. Full, evict, collision
     * 4. Full, evict, no collision
     */
    final int headIdx = hash(value);
    if (array[headIdx].value != -1) { // collision
      int free = -1;
      if (count == array.length) {   // array full, need to evict
        free = getAddressToEvict();
        Entry evict = array[free];
        LOG.finest("Array full. Evict candidate:" + evict);
        long evictValue = evict.value;
        if (evict.next != -1) {
          /**
           * When entry to be evicted is the head of a non-empty chain,
           * if it is simply evicted, the chain will be orphaned. Move the next
           * value in the chain on up, and free the replacement head's previous
           * address instead.
           *
           * So, the original head's value gets evicted, but its the second-in-line's
           * address which is used as evictIdx.
           *
           * If the chain had only one entry (which is also its head),
           * then its ok to replace it.
           */
          if (evict.head) {
            LOG.finest("Evict candidate is head");
            Entry head = evict;   // head can't be evicted!
            free = head.next;     // switch-a-roo
            evict = array[free];
            // copy completely next-in-line to head
            head.value = evict.value;
            head.next = evict.next;
            head.clockbit = evict.clockbit;
          } else {
            LOG.finest("Evict candidate not head. Linking up for eviction..");
            // link up the preceding and following items so evict can be removed
            Entry pred = getPredecessor(evict);
            pred.next = evict.next;
          }
        }

        try {
          LOG.finest("Purging idx=" + free + " value=" + evictValue);
          purger.purge(this, free, reconstitute(headIdx, evictValue));
        } catch (IOException e) {
          e.printStackTrace();
        }

      } else {  // there's still free space. let's find it
        free = headIdx;
        while (array[free].value != -1) {
          if (free < array.length - 1) {
            ++free;
          } else {
            free = 0;
          }
        }
//        LOG.finest("Found free address at " + free + "/" + array.length);
        ++count;
      }

      // new objects get inserted at the tail of the chain
      if (array[headIdx].next != -1) {    // chain with >1 entry
        Entry last = getPredecessor(array[headIdx]);
        last.next = free;
      } else {
        array[headIdx].next = free;
      }
      array[free].next = headIdx;         // circular
      array[free].head = false;
      array[free].clockbit = false;
      array[free].value = (int) value;    // store only the 32 least sig. bits
//      LOG.finest(array[free].toString());
    } else {
      array[headIdx].head = true;         // no collision, so must be head
      array[headIdx].value = (int) value; // store only the 32 least sig. bits
      ++count;
    }
    return headIdx;
  }

  /**
   * Entry MUST have a predecessor, i.e. e.next != -1
   *
   * @param e
   * @return
   */
  private Entry getPredecessor(Entry e) {
    Entry pred;
    for (pred = array[e.next]; ;) {
      if (pred.next == -1) break;
      if (array[pred.next] == e) break;  // found predecessor
      pred = array[pred.next];
    }
    return pred;
  }

  public int find(long obj) {
    int second = (int) obj; // get only the last 32 sig. bits
    boolean passedHead = false;
    for (int probe = hash(obj); probe != -1; probe = array[probe].next) {
      Entry entry = array[probe];
      if (entry.head) {
        if (passedHead) break;
        passedHead = true;
      }
      if (second == entry.value) {
        entry.clockbit = true;    // the clock bit
        return probe;
      }
    }
    return -1;
  }

  /**
   * Use 32 most sig. bits as hash. Lampson's abbreviated keys technique.
   *
   * @param obj
   * @return
   */
  protected int hash(long obj) {
    return (int) (obj >>> 48);  // length of value in bits (64) - size of cache in bits (16)
  }

  /**
   * This method has side-effect of incrementing clockIndex.
   *
   * @return
   */
  protected int getAddressToEvict() {
    LOG.finest("Finding address to evict..");
    int result = -1;
    while (true) {
      Entry e = array[clockIndex];
      if (!e.clockbit) {  // found victim
        result = clockIndex;
        if (clockIndex < array.length - 1)
          ++clockIndex;
        else {
          clockIndex = 0;
        }
        break;
      } else {
        e.clockbit = false;
      }
      if (clockIndex < array.length - 1)
        ++clockIndex;
      else {
        clockIndex = 0;
      }
    }
    LOG.finest("Evict candidate:" + result);
    return result;
  }

  /**
   * Reconstitute the full long given an index.
   *
   * @param idx
   * @return
   */
  public long reconstitute(int idx, long value) {
    return (idx << 48) | value;
  }

  public boolean contains(long chksm) {
    return find(chksm) == -1;
  }

  public static final class Entry {
    long value = -1;
    int next = -1;
    boolean clockbit = false;
    boolean head = false;

    void purge() {
      value = -1;
      next = -1;
      clockbit = false;
      head = false;
    }

    public boolean equals(Object o) {
      if (this == o) return true;
      if (!(o instanceof Entry)) return false;

      final Entry entry = (Entry) o;

      if (clockbit != entry.clockbit) return false;
      if (next != entry.next) return false;
      if (head != entry.head) return false;
      if (value != entry.value) return false;

      return true;
    }

    public String toString() {
      return "Entry{" +
          "object=" + value +
          ", next=" + next +
          ", head=" + head +
          ", mask=" + clockbit +
          "}";
    }
  }
}
