package org.supermind.crawl.util;

import org.apache.nutch.fs.NutchFileSystem;
import org.apache.nutch.io.*;
import org.apache.nutch.util.LogFormatter;

import java.io.File;
import java.io.IOException;
import java.util.Comparator;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Helper class to simplify interaction with a {@link MapFile}. Because a
 * MapFile is optimal in batch-update scenarios, an in-memory buffer is used
 * to perform batch updates. This buffer also improves performance when
 * read/updates have some locality.
 */
public abstract class MapFilePersister <K, V> {
  static Logger LOG =
      LogFormatter.getLogger(MapFilePersister.class.getName());

  /**
   * Buffer size.
   */
  int maxBufferSize = 5000;
  NutchFileSystem nfs;
//  String sorted;
//  File sortedFile;
  String mapdir;
  String baseDir;
  File mapdirFile;
  MapFile.Reader reader;

//  protected SequenceFile.Sorter sorter;

  /**
   * Current buffer position.
   */
  int bufferIdx = 0;

  /**
   * Buffer.
   */
  protected TreeMap<K, V> buffer = new TreeMap(getTypeComparator());

  /**
   * Should existing files/directories be overwritten?
   */
  boolean overwrite;

  /**
   * Initialize resources.
   *
   * @throws IOException
   */
  public void init() throws IOException {
    this.mapdirFile = new File(baseDir, Thread.currentThread().getName());
    this.mapdir = mapdirFile.toString();

//    this.sortedFile = new File(baseDir, "sort-" +
//        Thread.currentThread().getName() + ".tmp");
//    this.sorted = sortedFile.toString();

    if (overwrite) {
      nfs.delete(mapdirFile);
      nfs.delete(new File(mapdir + "-tmp"));
//      nfs.delete(sortedFile);
    }
    LOG.fine("MapFile initialized in " + mapdir);
  }

  /**
   * Return a new instance of the value. Defaults to {@link NullWritable#get}.
   *
   * @return
   */
  protected Writable getValueInstance(V v, boolean throwaway) {
    return NullWritable.get();
  }

  /**
   * Return a new instance of the key.
   *
   * @return
   */
  protected abstract WritableComparable getKeyInstance(K k, boolean throwaway);

  /**
   * Get comparator for MapFile key class.
   *
   * @return
   */
  protected abstract WritableComparator getKeyComparator();

  /**
   * Get comparator for type.
   *
   * @return
   */
  protected abstract Comparator<K> getTypeComparator();

  /**
   * Flush the buffer to disk. Buffer entries are appended to the tmpfile, which
   * is then sorted. A MapFile is created from the sorted file, so that fast
   * access to the entries is possible.
   *
   * @throws java.io.IOException
   */
//  public void flushToDisk() throws IOException {
//    long start = System.currentTimeMillis();
//    WritableComparable key = getKeyInstance();
//    Writable value = getValueInstance();
//
//    long checkpoint = System.currentTimeMillis();
//    if (tmpWriter == null) initTmpWriter();
//
//    writeBufferToTmp();               // write sorted values in buffer to tmpfile
//    if (LOG.isLoggable(Level.FINE))
//      LOG.fine("Took " + (System.currentTimeMillis() - checkpoint) + " ms to write buff=>tmp.");
//    checkpoint = System.currentTimeMillis();
//
//    tmpWriter.close();
//    tmpWriter = null;
//
//    if (sortedFile.exists()) {            // merge sort existing values with buffer values
//      File prevsorted = new File(this.sorted + ".old");
//      prevsorted.delete();
//      sortedFile.renameTo(prevsorted);
//      sorter.merge(new String[]{prevsorted.toString(), tmp}, this.sorted);
//    } else {
//      sorter.sort(tmp, this.sorted); // sort
//    }
//
//    if (LOG.isLoggable(Level.FINE))
//      LOG.fine("Took " + (System.currentTimeMillis() - checkpoint) + " ms to sort.");
//    checkpoint = System.currentTimeMillis();
//
//    MapFile.Writer mapwriter = null;
//    if (reader != null) { // close the reader, because it will prevent mapdir from being deleted
//      reader.close();
//      reader = null;
//    }
//
//    nfs.delete(mapdirFile);
//    // write mapdir from sorted values
//    try {
//      mapwriter = new MapFile.Writer(nfs, mapdir,
//          getMapFileKeyClass(), getMapFileValueClass());
//      SequenceFile.Reader sortReader = new SequenceFile.Reader(nfs, this.sorted);
//      while (sortReader.next(key, value)) {
//        mapwriter.append(key, value);
//      }
//      sortReader.close();
//    } finally {
//      if (mapwriter != null) mapwriter.close();
//      buffer.clear();
//    }
//    if (LOG.isLoggable(Level.FINE))
//      LOG.fine("Took " + (System.currentTimeMillis() - checkpoint) + " ms to write mapdir.");
//    checkpoint = System.currentTimeMillis();
//
//    long stop = System.currentTimeMillis();
//    if (LOG.isLoggable(Level.FINE))
//      LOG.fine("Took " + (stop - start) + " ms to flush buffer.");
//  }

  public void flushToDisk() throws IOException {
    LOG.info("Flushing buffer thread:" + Thread.currentThread().getName());
    long start = System.currentTimeMillis();
    WritableComparable key = getKeyInstance(null, false);
    Writable value = getValueInstance(null, false);
    final WritableComparator keyComparator = getKeyComparator();

//    if (sorter == null) {
//      sorter = new SequenceFile.Sorter(nfs,
//          keyComparator, value.getClass());
//    }

    long checkpoint = System.currentTimeMillis();

    if (LOG.isLoggable(Level.FINE))
      LOG.fine("Took " + (System.currentTimeMillis() - checkpoint) + " ms to write buff=>tmp.");
    checkpoint = System.currentTimeMillis();

    MapFile.Writer mapwriter = null;
    if (reader != null) { // close the reader, because it will prevent mapdir from being deleted
      reader.close();
      reader = null;
    }

    final File tmp = new File(mapdir + "-tmp");
    nfs.delete(tmp);
//    nfs.rename(mapdirFile, tmp); FIXME: nfs.rename broken for directory renames?
    mapdirFile.renameTo(tmp);

    try {
      mapwriter = new MapFile.Writer(nfs, mapdir,
          key.getClass(), value.getClass());

      if (tmp.exists()) {            // merge existing values with buffer values
        MapFile.Reader existingValues = new MapFile.Reader(nfs, tmp.getPath());
        while (existingValues.next(key, value)) {
          while (buffer.size() > 0 &&
              keyComparator.compare(key, getKeyInstance(buffer.firstKey(), true)) > 0) {
            K first = buffer.firstKey();
            V firstV = buffer.remove(first);
            WritableComparable keyInstance = getKeyInstance(first, true);
            final Writable valueInstance = getValueInstance(firstV, true);
            mapwriter.append(keyInstance, valueInstance);
          }
          mapwriter.append(key, value);
        }
        existingValues.close();
        nfs.delete(tmp);
      }

      for (K k : buffer.keySet()) {
        final WritableComparable keyInstance = getKeyInstance(k, true);
        final Writable valueInstance = getValueInstance(buffer.get(k), true);
        mapwriter.append(keyInstance, valueInstance);
      }

      if (LOG.isLoggable(Level.FINE))
        LOG.fine("Took " + (System.currentTimeMillis() - checkpoint) + " ms to sort.");
      checkpoint = System.currentTimeMillis();

    } finally {
      if (mapwriter != null) mapwriter.close();
      buffer.clear();
    }
    if (LOG.isLoggable(Level.FINE))
      LOG.fine("Took " + (System.currentTimeMillis() - checkpoint) + " ms to write mapdir.");
    checkpoint = System.currentTimeMillis();

    long stop = System.currentTimeMillis();
    if (LOG.isLoggable(Level.FINE))
      LOG.fine("Took " + (stop - start) + " ms to flush buffer.");

    LOG.info("Done flushing thread:" + Thread.currentThread().getName());

  }

  /**
   * Close resources.
   *
   * @throws IOException
   */
  public void close() throws IOException {
    if (buffer.size() > 0) {
      flushToDisk();
    }
    if (reader != null) {
      reader.close();
      reader = null;
    }
  }

  /**
   * Add a key/value pair.
   *
   * @param k
   * @param v
   * @throws IOException
   */
  public void add(K k, V v) throws IOException {
    if (bufferIdx >= maxBufferSize) {
      flushToDisk();
      bufferIdx = 0;
    }

    buffer.put(k, v);
    bufferIdx++;
  }

  /**
   * Set location of directory where the {@link MapFile} will be created.
   *
   * @param dir
   */
  public void setBaseDir(String dir) {
    this.baseDir = dir;
  }

  /**
   * Setter whether existing files/directories should be overwritten.
   *
   * @param overwrite
   */
  public void setOverwrite(boolean overwrite) {
    this.overwrite = overwrite;
  }

  /**
   * Set NutchFileSystem.
   *
   * @param nfs
   */
  public void setNfs(NutchFileSystem nfs) {
    this.nfs = nfs;
  }

  public void setMaxBufferSize(int maxBufferSize) {
    this.maxBufferSize = maxBufferSize;
  }
}
